"""POD file reading.

Reads L1b GAC/LAC data from POD series of satellites (NOAA-14 and earlier).
Format specification can be found in chapters 2 & 3 of the `POD user guide`_.

.. _POD user guide:
    https://www.ncei.noaa.gov/pub/data/satellite/publications/podguides/TIROS-N%20thru%20N-14/
"""

import datetime
import logging
import warnings
from enum import IntFlag

import numpy as np

from pygac.clock_offsets_converter import get_offsets
from pygac.correct_tsm_issue import TSM_AFFECTED_INTERVALS_POD, get_tsm_idx
from pygac.reader import DecodingError, NoTLEData, Reader, ReaderError
from pygac.slerp import slerp
from pygac.utils import file_opener

LOG = logging.getLogger(__name__)


class POD_QualityIndicator(IntFlag):
    """Quality Indicators.

    Source:
        POD guide Table 3.1.2.1-2. Format of quality indicators.
    """

    # POD guide Table 3.1.2.1-2. Format of quality indicators.
    FATAL_FLAG = 2**31  # Data should not be used for product generation
    TIME_ERROR = 2**30  # A time sequence error was detected while Processing
    # this frame
    DATA_GAP = 2**29  # A gap precedes this frame
    DATA_JITTER = 2**28  # Resync occurred on this frame
    CALIBRATION = 2**27  # Insufficient data for calibration
    NO_EARTH_LOCATION = 2**26  # Earth location data not available
    ASCEND_DESCEND = 2**25  # AVHRR Earth location indication of Ascending (=0)
    # or descending (=1) data
    PSEUDO_NOISE = 2**24  # Pseudo Noise (P/N) occurred (=1) on the frame,
    # data not used for calibration computations
    BIT_SYNC_STATUS = 2**23  # Drop lock during frame
    SYNC_ERROR = 2**22  # Frame Sync word error greater than zero
    FRAME_SYNC_LOCK = 2**21  # Frame Sync previously dropped lock
    FLYWHEELING = 2**20  # Flywheeling detected during this frame
    BIT_SLIPPAGE = 2**19  # Bit slippage detected during this frame
    # Solar blackbody contamination indicator
    # 0 = no correction
    # 1 = solar contamination corrected
    CH_3_CONTAMINATION = 2**18  # Channel 3 solar blackbody contamination
    CH_4_CONTAMINATION = 2**17  # Channel 4 solar blackbody contamination
    CH_5_CONTAMINATION = 2**16  # Channel 5 solar blackbody contamination
    # TIP Parity
    TIP_PARITY_1 = 2**15  # In first minor frame
    TIP_PARITY_2 = 2**14  # In second minor frame
    TIP_PARITY_3 = 2**13  # In third minor frame
    TIP_PARITY_4 = 2**12  # In fourth minor frame
    TIP_PARITY_5 = 2**11  # In fifth minor frame
    # Note: Bit 10 to 8, and 1 to 0 are spare bits. 7 to 2 define
    #       "SYNC ERRORS - Number of bit errors in frame sync" (6 bit integer?)


# common header
header0 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, ))])


# For L1B data before September 8, 1992
# http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/k/app-k.htm
header1 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("spare1", ">i1", (5, )),
                    ("data_set_name", "S44")])

# For L1B data between October 21, 1992 to November 15, 1994
# http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/l/app-l.htm
header2 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("spare1", ">i1", (5, )),
                    ("data_set_name", "S42"),
                    ("blankfill", "S2"),
                    ("julian_year_of_epoch", ">u2"),
                    ("julian_day_of_epoch", ">u2"),
                    ("millisecond_utc_epoch_time_of_day", ">u4"),
                    # Keplerian orbital elements
                    ("semi_major_axis", ">f8"),
                    ("eccentricity", ">f8"),
                    ("inclination", ">f8"),
                    ("argument_of_perigee", ">f8"),
                    ("right_ascension", ">f8"),
                    ("mean_anomaly", ">f8"),
                    # cartesian inertial true date of elements
                    ("x_component_of_position_vector", ">f8"),
                    ("y_component_of_position_vector", ">f8"),
                    ("z_component_of_position_vector", ">f8"),
                    ("x_dot_component_of_position_vector", ">f8"),
                    ("y_dot_component_of_position_vector", ">f8"),
                    ("z_dot_component_of_position_vector", ">f8")])

# For L1B data post November 15, 1994
# http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/podug/html/c2/sec2-0.htm
header3 = np.dtype([("noaa_spacecraft_identification_code", ">u1"),
                    ("data_type_code", ">u1"),
                    ("start_time", ">u2", (3, )),
                    ("number_of_scans", ">u2"),
                    ("end_time", ">u2", (3, )),
                    ("processing_block_id", "S7"),
                    ("ramp_auto_calibration", ">u1"),
                    ("number_of_data_gaps", ">u2"),
                    ("dacs_quality", ">u1", (6, )),
                    ("calibration_parameter_id", ">i2"),
                    ("dacs_status", ">u1"),
                    ("reserved_for_mounting_and_fixed_attitude_correction_indicator",
                     ">i1"),
                    ("nadir_earth_location_tolerance", ">i1"),
                    ("spare1", ">i1"),
                    ("start_of_data_set_year", ">u2"),
                    ("data_set_name", "S44"),
                    ("year_of_epoch", ">u2"),
                    ("julian_day_of_epoch", ">u2"),
                    ("millisecond_utc_epoch_time_of_day", ">u4"),
                    # Keplerian orbital elements
                    ("semi_major_axis", ">i4"),
                    ("eccentricity", ">i4"),
                    ("inclination", ">i4"),
                    ("argument_of_perigee", ">i4"),
                    ("right_ascension", ">i4"),
                    ("mean_anomaly", ">i4"),
                    # cartesian inertial true date of elements
                    ("x_component_of_position_vector", ">i4"),
                    ("y_component_of_position_vector", ">i4"),
                    ("z_component_of_position_vector", ">i4"),
                    ("x_dot_component_of_position_vector", ">i4"),
                    ("y_dot_component_of_position_vector", ">i4"),
                    ("z_dot_component_of_position_vector", ">i4"),
                    # future use
                    ("yaw_fixed_error_correction", ">i2"),
                    ("roll_fixed_error_correction", ">i2"),
                    ("pitch_fixed_error_correction", ">i2")])

# archive header
tbm_header = np.dtype([("fill", "S30"),
                       ("data_set_name", "S44"),
                       ("select_flag", "S1"),
                       ("beginning_latitude", "S3"),
                       ("ending_latitude", "S3"),
                       ("beginning_longitude", "S4"),
                       ("ending_longitude", "S4"),
                       ("start_hour", "S2"),
                       ("start_minute", "S2"),
                       ("number_of_minutes", "S3"),
                       ("appended_data_flag", "S1"),
                       ("channel_select_flag", "S1", (20, )),
                       ("sensor_data_word_size", "S2"),
                       ("fill2", "S3")])


class PODReader(Reader):
    """The POD reader."""

    spacecrafts_orbital = {25: "tiros n",
                           2: "noaa 6",
                           4: "noaa 7",
                           6: "noaa 8",
                           7: "noaa 9",
                           8: "noaa 10",
                           1: "noaa 11",
                           5: "noaa 12",
                           3: "noaa 14",
                           }
    spacecraft_names = {25: "tirosn",
                        2: "noaa6",
                        4: "noaa7",
                        6: "noaa8",
                        7: "noaa9",
                        8: "noaa10",
                        1: "noaa11",
                        5: "noaa12",
                        3: "noaa14",
                        }

    tsm_affected_intervals = TSM_AFFECTED_INTERVALS_POD

    QFlag = POD_QualityIndicator
    _quality_indicators_key = "quality_indicators"

    def correct_scan_line_numbers(self):
        """Correct the scan line numbers."""
        # Perform common corrections first.
        results = super().correct_scan_line_numbers()

        # cleaning up the data
        min_scanline_number = np.amin(
            np.absolute(self.scans["scan_line_number"][:]))
        if self.scans["scan_line_number"][0] == self.scans["scan_line_number"][-1] + 1:
            while self.scans["scan_line_number"][0] != min_scanline_number:
                self.scans = np.roll(self.scans, -1)
        else:
            while self.scans["scan_line_number"][0] != min_scanline_number:
                self.scans = self.scans[1:]

        self.scans = self.scans[self.scans["scan_line_number"] != 0]
        return results

    def read(self, filename, fileobj=None):
        """Read the data.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)

        Returns:
            header: numpy record array
                The header metadata
            scans: numpy record array
                The scanlines

        """
        self.filename = filename
        LOG.info("Reading %s", self.filename)
        # choose the right header depending on the date
        with file_opener(fileobj or filename) as fd_:
            self.tbm_head, self.head = self.read_header(
                filename, fileobj=fd_, header_date=self.header_date)
            if self.tbm_head:
                tbm_offset = tbm_header.itemsize
            else:
                tbm_offset = 0
            # read scan lines until end of file
            fd_.seek(self.offset + tbm_offset, 0)
            buffer = fd_.read()
            buffer = self._truncate_padding_record(buffer)
            count = self.head["number_of_scans"]
            self._read_scanlines(buffer, count)
        year, jday, _ = self.decode_timestamps(self.head["start_time"])
        start_date = (datetime.date(year, 1, 1) +
                      datetime.timedelta(days=int(jday) - 1))
        if self.correct_scanlines:
            self.correct_scan_line_numbers()
        self.spacecraft_id = self.head["noaa_spacecraft_identification_code"]
        if self.spacecraft_id == 1 and start_date < datetime.date(1982, 1, 1):
            self.spacecraft_id = 25
        self.spacecraft_name = self.spacecraft_names[self.spacecraft_id]
        LOG.info(
            "Reading %s data", self.spacecrafts_orbital[self.spacecraft_id])
        return self.head, self.scans

    def _truncate_padding_record(self, buffer):
        """Remove the padding record (if present) from the data stream.

        In POD files each scanline was stored as a "logical" record, and these
        were written to the tapes in "physical" records. In the case of GAC
        data there were two logical records (scanlines) per physical record,
        so an odd number of scanlines would result in the final physical record
        containing only one scanline and one "logical" record of padding.
        """
        scanlines_per_record = self.offset // self.scanline_type.itemsize
        # Do nothing if the logical and physical records are the same size (LAC).
        if scanlines_per_record == 1:
            return buffer
        # File should contain an integer number of physical records
        n_logical = len(buffer) // self.scanline_type.itemsize
        if n_logical % scanlines_per_record != 0:
            LOG.warning("Unexpected record length for POD file (incomplete physical record?)")
            warnings.warn("Unexpected record length for POD file (incomplete physical record?)",
                          category=RuntimeWarning, stacklevel=2)
        # How many physical / logical records do we expect based on the file header?
        expected_scanlines = int(self.head["number_of_scans"])
        expected_physical = (expected_scanlines + scanlines_per_record - 1) // scanlines_per_record
        expected_logical = expected_physical * scanlines_per_record
        # Only trim the padding if the file is the expected size, so any unexpected
        # cases will still be dealt with in _read_scanlines
        if n_logical == expected_logical and n_logical > expected_scanlines:
            buffer = buffer[:expected_scanlines * self.scanline_type.itemsize]
        return buffer

    @classmethod
    def read_header(cls, filename, fileobj=None, header_date="auto"):
        """Read the file header.

        Args:
            filename (str): Path to GAC/LAC file
            fileobj: An open file object to read from. (optional)
            header_date: date to use to choose the header. Defaults to "auto"
                to use the data to pick the header corresponding to the date of the file.

        Returns:
            archive_header (struct): archive header
            header (struct): file header

        """
        # choose the right header depending on the date
        with file_opener(fileobj or filename) as fd_:
            # read tbm_header if present
            _tbm_head, = np.frombuffer(
                fd_.read(tbm_header.itemsize),
                dtype=tbm_header, count=1)
            try:
                tbm_head = cls._validate_tbm_header(_tbm_head)
                tbm_offset = tbm_header.itemsize
            except DecodingError:
                tbm_head = None
                tbm_offset = 0

            fd_.seek(tbm_offset, 0)
            header = cls.choose_header_based_on_timestamp(header_date, fd_)
            fd_.seek(tbm_offset, 0)
            # need to copy frombuffer to have write access on head
            head, = np.frombuffer(
                fd_.read(header.itemsize),
                dtype=header, count=1).copy()
        head = cls._correct_data_set_name(head, filename)
        cls._validate_header(head)
        return tbm_head, head

    @classmethod
    def _validate_tbm_header(cls, potential_tbm_header):
        data_set_name = potential_tbm_header["data_set_name"]
        allowed_empty = (42*b"\x00" + b"  ")
        if data_set_name == allowed_empty:
            return potential_tbm_header.copy()

        # This will raise a DecodingError if the data_set_name is not valid.
        cls._decode_data_set_name(data_set_name)
        return potential_tbm_header.copy()


    @classmethod
    def choose_header_based_on_timestamp(cls, header_date, fd_):
        """Choose the header dtype based on the timestamp."""
        if header_date == "auto":
            header_date = cls.get_start_date(fd_)
        if header_date < datetime.date(1992, 9, 8):
            header = header1
        elif header_date <= datetime.date(1994, 11, 15):
            header = header2
        else:
            header = header3
        return header

    @classmethod
    def get_start_date(cls, fd_):
        """Get the start time from the filestream."""
        head0, = np.frombuffer(
                    fd_.read(header0.itemsize),
                    dtype=header0, count=1)
        year, jday, _ = cls.decode_timestamps(head0["start_time"])
        header_date = (datetime.date(year, 1, 1) +
                       datetime.timedelta(days=int(jday) - 1))

        return header_date

    @classmethod
    def _validate_header(cls, header):
        """Check if the header belongs to this reader."""
        # call super to enter the Method Resolution Order (MRO)
        super(PODReader, cls)._validate_header(header)
        LOG.debug("validate header")
        data_set_name = header["data_set_name"].decode()
        # split header into parts
        creation_site, transfer_mode, platform_id = (
            data_set_name.split(".")[:3])
        allowed_ids = ["TN", "NA", "NB", "NC", "ND", "NE", "NF", "NG",
                       "NH", "NI", "NJ"]
        if platform_id not in allowed_ids:
            raise ReaderError('Improper platform id "%s"!' % platform_id)

    def get_header_timestamp(self):
        """Get the timestamp from the header.

        Returns:
            A datetime object containing the timestamp from the header.

        Raises:
            A ValueError if the timestamp is corrupt.

        """
        year, jday, msec = self.decode_timestamps(self.head["start_time"])
        try:
            return self.to_datetime(self.to_datetime64(year=year, jday=jday,
                                                       msec=msec))
        except ValueError as err:
            raise ValueError("Corrupt header timestamp: {0}".format(err))

    @staticmethod
    def decode_timestamps(encoded):
        """Decode timestamps.

        Returns:
            year
            day of year
            milliseconds since 00:00

        """
        ndims = len(encoded.shape)
        if ndims == 1:
            # Single header timestamp
            enc0 = encoded[0]
            enc1 = encoded[1]
            enc2 = encoded[2]
        elif ndims == 2:
            # Scanline timestamps
            enc0 = encoded[:, 0]
            enc1 = encoded[:, 1]
            enc2 = encoded[:, 2]
        else:
            raise ValueError("Invalid timestamp dimension")

        year = enc0 >> 9
        year = np.where(year > 75, year + 1900, year + 2000)
        jday = (enc0 & 0x1FF)
        msec = ((np.uint32(enc1 & 2047) << 16) | (np.uint32(enc2)))

        return year, jday, msec

    def _get_times_from_file(self):
        return self.decode_timestamps(self.scans["time_code"])

    def compute_clock_offsets(self):
        error_utcs, clock_error = get_offsets(self.spacecraft_name)

        error_utcs = np.array(error_utcs, dtype="datetime64[ms]")
        # interpolate to get the clock offsets at the scan line utcs
        # the clock_error is given in seconds, so offsets are in seconds, too.
        offsets = np.interp(self._times_as_np_datetime64.astype(np.uint64),
                            error_utcs.astype(np.uint64),
                            clock_error)
        return offsets

    def _adjust_clock_drift(self):
        """Adjust the geolocation to compensate for the clock error."""
        tic = datetime.datetime.now()
        self.get_times()
        try:
            offsets = self.compute_clock_offsets()

        except KeyError:
            LOG.info("No clock drift info available for %s",
                     self.spacecraft_name)
            return
        LOG.info("Adjusting for clock drift of %s to %s seconds.",
                 str(min(offsets)),
                 str(max(offsets)))

        # For the interpolation of geolocations, we need to prepend/append
        # lines. The conversion from offset to line numbers is given by the
        # scan rate = 1/scan frequency.
        scan_rate = datetime.timedelta(milliseconds=1/self.scan_freq)
        offset_lines = offsets / scan_rate.total_seconds()

        # To avoid trouble with missing lines, we will construct an
        # array that covers the whole interpolation range.
        scan_lines = self.scans["scan_line_number"]
        shifted_lines = scan_lines - offset_lines
        shifted_lines_floor = np.floor(shifted_lines).astype(int)
        # compute the line range, note that the max requires a "+1"
        # to cover the interpolation range because of the floor.
        min_line = min(scan_lines.min(), shifted_lines_floor.min())
        max_line = max(scan_lines.max(), shifted_lines_floor.max()+1)
        num_lines = max_line - min_line + 1
        missed_lines = np.setdiff1d(np.arange(min_line, max_line+1), scan_lines)
        missed_utcs = ((missed_lines - scan_lines[0])*np.timedelta64(scan_rate, "ms")
                       + self._times_as_np_datetime64[0])
        # calculate the missing geo locations
        try:
            missed_lons, missed_lats = self._compute_lonlat_from_tles(missed_utcs)
        except NoTLEData as err:
            LOG.warning("Cannot perform clock drift correction: %s", str(err))
            return

        # create arrays of lons and lats for interpolation. The locations
        # correspond to not yet corrected utcs, i.e. the time difference from
        # one line to the other should be equal to the scan rate.
        pixels_per_line = self.lats.shape[1]
        complete_lons = np.full((num_lines, pixels_per_line), np.nan,
                                dtype=np.float64)
        complete_lats = np.full((num_lines, pixels_per_line), np.nan,
                                dtype=np.float64)

        complete_lons[scan_lines - min_line] = self.lons
        complete_lats[scan_lines - min_line] = self.lats
        complete_lons[missed_lines - min_line] = missed_lons
        complete_lats[missed_lines - min_line] = missed_lats

        # perform the slerp interpolation to the corrected utc times
        slerp_t = shifted_lines - shifted_lines_floor  # in [0, 1)
        slerp_res = slerp(complete_lons[shifted_lines_floor - min_line, :],
                          complete_lats[shifted_lines_floor - min_line, :],
                          complete_lons[shifted_lines_floor - min_line + 1, :],
                          complete_lats[shifted_lines_floor - min_line + 1, :],
                          slerp_t[:, np.newaxis, np.newaxis])

        # set corrected values
        self.lons = slerp_res[:, :, 0]
        self.lats = slerp_res[:, :, 1]
        self._times_as_np_datetime64 -= (offsets * 1000).astype("timedelta64[ms]")

        toc = datetime.datetime.now()
        LOG.debug("clock drift adjustment took %s", str(toc - tic))

    def _get_lonlat_from_file(self):
        lats = self.scans["earth_location"]["lats"] / np.float32(128.0)
        lons = self.scans["earth_location"]["lons"] / np.float32(128.0)
        return lons, lats

    def get_telemetry(self):
        """Get the telemetry.

        Returns:
            prt_counts: np.array
            ict_counts: np.array
            space_counts: np.array

        """
        number_of_scans = self.scans["telemetry"].shape[0]
        decode_tele = np.zeros((int(number_of_scans), 105))
        decode_tele[:, ::3] = (self.scans["telemetry"] >> 20) & 1023
        decode_tele[:, 1::3] = (self.scans["telemetry"] >> 10) & 1023
        decode_tele[:, 2::3] = self.scans["telemetry"] & 1023

        prt_counts = np.mean(decode_tele[:, 17:20], axis=1)

        # getting ICT counts

        ict_counts = np.zeros((int(number_of_scans), 3))
        ict_counts[:, 0] = np.mean(decode_tele[:, 22:50:3], axis=1)
        ict_counts[:, 1] = np.mean(decode_tele[:, 23:51:3], axis=1)
        ict_counts[:, 2] = np.mean(decode_tele[:, 24:52:3], axis=1)

        # getting space counts

        space_counts = np.zeros((int(number_of_scans), 3))
        space_counts[:, 0] = np.mean(decode_tele[:, 54:100:5], axis=1)
        space_counts[:, 1] = np.mean(decode_tele[:, 55:101:5], axis=1)
        space_counts[:, 2] = np.mean(decode_tele[:, 56:102:5], axis=1)

        return prt_counts, ict_counts, space_counts

    @staticmethod
    def _get_ir_channels_to_calibrate():
        ir_channels_to_calibrate = [3, 4, 5]
        return ir_channels_to_calibrate

    def postproc(self, channels):
        """No POD specific postprocessing to be done."""
        pass

    def get_tsm_pixels(self, channels):
        """Determine pixels affected by the scan motor issue.

        Uses channels 1, 2, 4 and 5. Neither 3a, nor 3b.
        """
        return get_tsm_idx(channels[:, :, 0], channels[:, :, 1],
                           channels[:, :, 3], channels[:, :, 4])

    def _get_calibrated_channels_uniform_shape(self):
        """Prepare the channels as input for gac_io.save_gac."""
        _channels = self.get_calibrated_channels()
        # prepare input
        # maybe there is a better (less memory requiring) method
        shape = _channels.shape[:-1] + (6,)
        # empty_like does not know the kwarg shape in older versions
        channels = np.empty(shape, dtype=_channels.dtype)
        channels[:, :, 0] = _channels[:, :, 0]
        channels[:, :, 1] = _channels[:, :, 1]
        channels[:, :, 2] = np.nan
        channels[:, :, 3] = _channels[:, :, 2]
        channels[:, :, 4] = _channels[:, :, 3]
        channels[:, :, 5] = _channels[:, :, 4]
        return channels


def main_pod(reader_cls, filename, start_line, end_line):
    """Generate a l1c file."""
    tic = datetime.datetime.now()
    reader = reader_cls.fromfile(filename)
    reader.save(int(start_line), int(end_line))
    LOG.info("pygac took: %s", str(datetime.datetime.now() - tic))



"""Generic reader for GAC data.

Can't be used as is, has to be subclassed to add specific read functions.
"""

import logging
import warnings

import numpy as np

try:
    from pyorbital.geoloc_instrument_definitions import avhrr_gac_from_times
except ImportError:
    # In pyorbital 1.9.2 and earlier avhrr_gac returned LAC geometry
    from pyorbital.geoloc_instrument_definitions import avhrr_gac
    def avhrr_gac_from_times(times, points):
        return avhrr_gac(times, points*5+3.5)
    warnings.warn("pyorbital version does not support avhrr_gac_from_times. " +
                  "Computation of missing longitude/latitudes may be incorrect.")


from pygac.reader import Reader, ReaderError

LOG = logging.getLogger(__name__)


class GACReader(Reader):
    """Reader for GAC data."""

    # Scanning frequency (scanlines per millisecond)
    scan_freq = 2.0 / 1000.0
    # Max scanlines
    max_scanlines = 15000
    lonlat_sample_points = np.arange(4, 405, 8)

    def __init__(self, *args, **kwargs):
        """Init the GAC reader."""
        super().__init__(*args, **kwargs)
        self.scan_width = 409
        self.geoloc_definition = avhrr_gac_from_times

    @classmethod
    def _validate_header(cls, header):
        """Check if the header belongs to this reader."""
        # call super to enter the Method Resolution Order (MRO)
        super(GACReader, cls)._validate_header(header)
        LOG.debug("validate header")
        data_set_name = header["data_set_name"].decode()
        # split header into parts
        creation_site, transfer_mode, platform_id = (
            data_set_name.split(".")[:3])
        if transfer_mode != "GHRR":
            raise ReaderError('Improper transfer mode "%s"!' % transfer_mode)



"""Test the GAC KLM reader."""

import datetime as dt
from unittest import mock

import numpy as np
import numpy.testing
import xarray as xr

from pygac.gac_klm import GACKLMReader
from pygac.klm_reader import header
from pygac.lac_klm import LACKLMReader, scanline
from pygac.tests.utils import CalledWithArray


class TestKLM:
    """Test the klm reader."""

    def setup_method(self):
        """Set up the tests."""
        self.reader = GACKLMReader()

    def test_get_lonlat(self):
        """Test readout of lon/lat coordinates."""
        lons_exp = np.array([[2, 4],
                             [6, 8]])
        lats_exp = np.array([[1, 3],
                             [5, 7]])

        self.reader.scans = {"earth_location": {"lats": lats_exp * 1e4,
                                                "lons": lons_exp * 1e4}}

        lons, lats = self.reader._get_lonlat_from_file()
        numpy.testing.assert_array_equal(lons, lons_exp)
        numpy.testing.assert_array_equal(lats, lats_exp)

    def test_get_header_timestamp(self):
        """Test readout of header timestamp."""
        self.reader.head = {
            "start_of_data_set_year": np.array([2019]),
            "start_of_data_set_day_of_year": np.array([123]),
            "start_of_data_set_utc_time_of_day": np.array([123456])
        }
        time = self.reader.get_header_timestamp()
        assert time == dt.datetime(2019, 5, 3, 0, 2, 3, 456000)

    def test_get_times(self):
        """Test readout of scanline timestamps."""
        self.reader.scans = {"scan_line_year": 1,
                             "scan_line_day_of_year": 2,
                             "scan_line_utc_time_of_day": 3}
        assert self.reader._get_times_from_file() == (1, 2, 3)

    def test_get_ch3_switch(self):
        """Test channel 3 identification."""
        self.reader.scans = {
            "scan_line_bit_field": np.array([1, 2, 3, 4, 5, 6])}
        switch_exp = np.array([1, 2, 3, 0, 1, 2])
        numpy.testing.assert_array_equal(
            self.reader.get_ch3_switch(), switch_exp)

    def test_postproc(self):
        """Test KLM specific postprocessing."""
        self.reader.scans = {
            "scan_line_bit_field": np.array([0, 1, 2])}
        channels = np.array([[[1., 2., 3., 4.],
                              [1., 2., 3., 4.]],
                             [[1., 2., 3., 4.],
                              [1., 2., 3., 4.]],
                             [[1., 2., 3, 4.],
                              [1., 2., 3, 4.]]])  # (lines, pixels, channels)

        masked_exp = np.array([[[1., 2., np.nan, 4.],
                                [1., 2., np.nan, 4.]],
                               [[1., 2., 3., np.nan],
                                [1., 2., 3., np.nan]],
                               [[1., 2., np.nan, np.nan],
                                [1., 2., np.nan, np.nan]]])
        channels = xr.DataArray(channels, dims=["scan_line_index", "columns", "channel_name"],
                                coords=dict(channel_name=["1", "2", "3a", "3b"] ))
        ds = xr.Dataset(dict(channels=channels))
        self.reader.postproc(ds)  # masks in-place
        numpy.testing.assert_array_equal(channels, masked_exp)

    def test_quality_indicators(self):
        """Test the quality indicator unpacking."""
        reader = self.reader
        QFlag = reader.QFlag

        dtype = np.uint32

        quality_indicators = np.array([
            0,  # nothing flagged
            np.iinfo(dtype).max,  # everything flagged
            QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION,
            QFlag.TIME_ERROR | QFlag.DATA_GAP,
            QFlag.FATAL_FLAG
        ], dtype=np.int64)
        reader.scans = {self.reader._quality_indicators_key: quality_indicators}
        # test mask, i.e. QFlag.FATAL_FLAG | QFlag.CALIBRATION | QFlag.NO_EARTH_LOCATION
        expected_mask = np.array([False, True, True, False, True], dtype=bool)
        numpy.testing.assert_array_equal(reader.mask, expected_mask)
        # test individual flags
        assert reader._get_corrupt_mask(flags=QFlag.FATAL_FLAG).any()
        # count the occurence (everything flagged and last entrance => 2)
        assert reader._get_corrupt_mask(flags=QFlag.FATAL_FLAG).sum() == 2


class TestGACKLM:
    """Tests for gac klm."""

    def setup_method(self):
        """Set up the tests."""
        self.reader = GACKLMReader()

    @mock.patch("pygac.klm_reader.get_tsm_idx")
    def test_get_tsm_pixels(self, get_tsm_idx):
        """Test channel set used for TSM correction."""
        ones = np.ones((409, 100))
        zeros = np.zeros(ones.shape)
        ch1 = 1*ones
        ch2 = 2*ones
        ch4 = 4*ones
        ch5 = 5*ones
        channels = np.dstack((ch1, ch2, zeros, zeros, ch4, ch5))
        self.reader.get_tsm_pixels(channels)
        get_tsm_idx.assert_called_with(CalledWithArray(ch1),
                                       CalledWithArray(ch2),
                                       CalledWithArray(ch4),
                                       CalledWithArray(ch5))


class TestLACKLM:
    """Tests for lac klm."""

    def setup_method(self):
        """Set up the tests."""
        self.reader = LACKLMReader()
        self.reader.scans = np.ones(100, dtype=scanline)
        self.reader.scans["scan_line_year"] = 2001
        self.reader.scans["sensor_data"] = 2047
        self.reader.head = np.ones(1, dtype=header)[0]
        self.reader.spacecraft_id = 12
        self.reader.head["noaa_spacecraft_identification_code"] = self.reader.spacecraft_id
        self.reader.spacecraft_name = "metopa"
        self.reader.scans["scan_line_number"] = np.arange(100)
        # PRT
        self.reader.scans["telemetry"]["PRT"] = 400
        self.reader.scans["telemetry"]["PRT"][0::5, :] = 0
        self.reader.tle_lines = ("first line", "second line")

    def test_get_ch3_switch(self):
        """Test channel 3 identification."""
        self.reader.scans = {
            "scan_line_bit_field": np.array([1, 2, 3, 4, 5, 6])}
        switch_exp = np.array([1, 2, 3, 0, 1, 2])
        numpy.testing.assert_array_equal(
            self.reader.get_ch3_switch(), switch_exp)

    def test_calibrate_channels(self):
        """Test channel calibration."""
        # ICT
        self.reader.scans["back_scan"] = 400
        self.reader.scans["back_scan"][0::5, :] = 0
        # Space
        self.reader.scans["space_data"] = 400
        self.reader.scans["space_data"][0::5, :] = 0

        assert np.any(np.isfinite(self.reader.get_calibrated_channels()))

    def test_calibrate_inactive_3b(self):
        """Test calibration of an inactive 3b."""
        channels = self.reader.get_calibrated_channels()
        assert np.all(np.isnan(channels[:, :, 3]))


def test_gac_scanline_dtype():
    """Test the gac scanline size."""
    from pygac.gac_klm import scanline
    assert scanline.itemsize == 4608

def test_lac_scanline_dtype():
    """Test the lac scanline size."""
    from pygac.lac_klm import scanline
    assert scanline.itemsize == 15872



"""Module configuration class.

Read and manage module configuration.
"""
import configparser
import logging
import os

LOG = logging.getLogger(__name__)


class Configuration(configparser.ConfigParser):
    """Configuration container for pygac."""

    config_file = ""

    def read(self, config_file):
        """Read and parse the configuration file

        Args:
            config_file (str): path to the config file

        Note:
            In contrast to the parent method, this implementation
            only accepts a single file and raises an exception if
            the given file is not readable.
        """
        if not os.path.isfile(config_file):
            raise FileNotFoundError(
                'Given config path "%s" is not a file!' % config_file)
        try:
            super().read(config_file)
        except configparser.NoSectionError:
            LOG.error('Failed reading configuration file: "%s"'
                      % config_file)
            raise
        self.config_file = config_file


_config = Configuration()


def get_config(initialized=True):
    """Retrun the module configuration.

    Args:
        initialized (bool): if true ensure that the configuration has
                            been initialized (default: True)
    """
    global _config
    if initialized and not _config.sections():
        LOG.info('Configuration has not been initialized. Use'
                 ' environment variable "PYGAC_CONFIG_FILE"')
        try:
            config_file = os.environ["PYGAC_CONFIG_FILE"]
        except KeyError:
            LOG.error("Environment variable PYGAC_CONFIG_FILE not set!")
            raise
        _config.read(config_file)
    return _config


def read_config_file(config_file):
    """Read a given config file."""
    config = get_config(initialized=False)
    config.read(config_file)



"""Calibration coefficients and generic calibration functions
"""
from __future__ import division

import datetime as dt
import hashlib
import json
import logging
import sys
import warnings
from collections import namedtuple
from enum import Enum
from importlib.resources import files

import numpy as np

LOG = logging.getLogger(__name__)


def calibrate(ds, custom_coeffs=None, coeffs_file=None):
    """Calibrate the dataset `ds`, possibly using custom_coeffs or a coeff file.

    Args:
        ds: xarray Dataset object containing the data to calibrate.
        custom_calibration: dictionary with a subset of user defined satellite specific
                            calibration coefficients
        calibration_file: path to json file containing default calibrations
    """
    channels = ds["channels"].data
    times = ds.coords["times"]
    scan_line_numbers = ds["scan_line_index"].data

    calibration_coeffs = Calibrator(
            ds.attrs["spacecraft_name"],
            custom_coeffs=custom_coeffs,
            coeffs_file=coeffs_file
        )

    # `times` is in nanosecond resolution. However, convertion from datetime64[ns] to datetime objects
    # does not work, and anyway only the date is needed.
    start_time = times[0].dt
    year = start_time.year.item()
    jday = start_time.dayofyear.item()


    corr = ds.attrs["sun_earth_distance_correction_factor"]

    # how many reflective channels are there ?
    tot_ref = channels.shape[2] - 3

    channels[:, :, 0:tot_ref] = calibrate_solar(
        channels[:, :, 0:tot_ref],
        np.arange(tot_ref),
        year, jday,
        calibration_coeffs,
        corr
    )

    prt = ds["prt_counts"].data
    ict = ds["ict_counts"].data
    space = ds["space_counts"].data

    ir_channels_to_calibrate = [3, 4, 5]

    for chan in ir_channels_to_calibrate:
        channels[:, :, chan - 6] = calibrate_thermal(
            channels[:, :, chan - 6],
            prt,
            ict[:, chan - 3],
            space[:, chan - 3],
            scan_line_numbers,
            chan,
            calibration_coeffs
        )

    new_ds = ds.copy()
    new_ds["channels"].data = channels

    new_ds.attrs["calib_coeffs_version"] = calibration_coeffs.version

    return new_ds


class CoeffStatus(Enum):
    """Indicates the status of calibration coefficients."""
    NOMINAL = "nominal"
    PROVISIONAL = "provisional"
    EXPERIMENTAL = "experimental"


class Calibrator:
    """Factory class to create namedtuples holding the calibration coefficients.

    Attributes:
        fields: coefficient names
        Calibrator: namedtuple constructor
        default_coeffs: dictonary containing default values for all spacecrafts
    """
    version_hashs = {
        "963af9b66268475ed500ad7b37da33c5": {
            "name": "PATMOS-x, v2017r1",
            "status": CoeffStatus.NOMINAL
        },
        "689386c822de18a07194ac7fd71652ea": {
            "name": "PATMOS-x, v2017r1, with provisional coefficients for MetOp-C",
            "status": CoeffStatus.PROVISIONAL
        },
        "e8735ec394ecdb87b7edcd261e72d2eb": {
            "name": "PATMOS-x, v2023",
            "status": CoeffStatus.PROVISIONAL
        },
    }
    fields = [
        "dark_count", "gain_switch", "s0", "s1", "s2", "b",  # "b0", "b1", "b2",
        "centroid_wavenumber", "space_radiance", "to_eff_blackbody_intercept",
        "to_eff_blackbody_slope", "date_of_launch", "d", "spacecraft", "version"
    ]

    Calibrator = namedtuple("Calibrator", fields)
    default_coeffs = None
    default_file = None
    default_version = None

    def __new__(cls, spacecraft, custom_coeffs=None, coeffs_file=None):
        """Creates a namedtuple for calibration coefficients of a given spacecraft

        Args:
            spacecraft (str): spacecraft name in pygac convention
            custom_coeffs (dict): custom coefficients (optional)
            coeffs_file (str): path to coefficents file (optional)

        Returns:
            calibrator (namedtuple): calibration coefficients

        Note:
            The coefficients in coeffs_file serve as default values if no custom_coeffs
            are given. If omitted, Calibrator uses the PyGAC internal defaults.
        """
        if cls.default_coeffs is None or cls.default_file != coeffs_file:
            cls.default_file = coeffs_file
            cls.default_coeffs, cls.default_version = cls.read_coeffs(coeffs_file)
        if custom_coeffs:
            LOG.info('Using following custom coefficients "%s".', custom_coeffs)
        customs = custom_coeffs or {}
        defaults = cls.default_coeffs[spacecraft]
        coeffs = defaults.copy()
        coeffs.update(customs)

        # transpose the coefficient order from channel - coeff to coeff - channel
        # and store as arrays for vectorized calls of calibration functions
        arraycoeffs = dict.fromkeys(cls.fields)
        # visible channels
        for key in ("dark_count", "gain_switch", "s0", "s1", "s2"):
            arraycoeffs[key] = np.array([
                coeffs[channel][key]
                for channel in ("channel_1", "channel_2", "channel_3a")
            ], dtype=float)
        # thermal channels
        for key in ("centroid_wavenumber", "space_radiance",
                    "to_eff_blackbody_intercept", "to_eff_blackbody_slope"):
            arraycoeffs[key] = np.array([
                coeffs[channel][key]
                for channel in ("channel_3b", "channel_4", "channel_5")
            ], dtype=float)
        arraycoeffs["b"] = np.array([
            [
                coeffs[channel][key]
                for key in ("b0", "b1", "b2")
            ]
            for channel in ("channel_3b", "channel_4", "channel_5")
        ], dtype=float)
        # thermometers
        # Note, that "thermometer_0" does not exists, and is filled with zeros to
        # account for the PRT reset every fifth scanline
        arraycoeffs["d"] = np.array([
            [
                coeffs.get("thermometer_{0}".format(t), {}).get("d{0}".format(d), 0.0)
                for t in range(5)
            ]
            for d in range(5)
        ], dtype=float)
        # parse date of launch
        date_of_launch_str = coeffs["date_of_launch"].replace("Z", "+00:00")
        if sys.version_info < (3, 7):
            # Note that here any time information is lost
            import dateutil.parser
            date_of_launch = dateutil.parser.parse(date_of_launch_str)
        else:
            # datetime.fromisoformat() was introduced in Python-3.7
            date_of_launch = dt.datetime.fromisoformat(
                date_of_launch_str).astimezone(dt.timezone.utc)
        # remove time zone information (easier to handle in calculations)
        arraycoeffs["date_of_launch"] = date_of_launch.replace(tzinfo=None)
        arraycoeffs["spacecraft"] = spacecraft
        arraycoeffs["version"] = cls.default_version
        if custom_coeffs:
            arraycoeffs["version"] = None
        # create namedtuple
        calibrator = cls.Calibrator(**arraycoeffs)
        return calibrator

    @classmethod
    def read_coeffs(cls, coeffs_file):
        """Read calibration coefficients for all satellites from file.

        Args:
            coeffs_file (str): path to coefficients file

        Returns:
            coeffs (dict): dictionary containing coefficients for all satellites
            version (str): version of the coefficients (None if unknown)
        """
        if coeffs_file:
            LOG.info('Read calibration coefficients from "%s"', coeffs_file)
        else:
            LOG.debug("Read PyGAC internal calibration coefficients.")
            coeffs_file = files("pygac") / "data/calibration.json"
        with open(coeffs_file, mode="rb") as json_file:
            content = json_file.read()
            coeffs = json.loads(content)
            version = cls._get_coeffs_version(content)
        return coeffs, version

    @classmethod
    def _get_coeffs_version(cls, coeff_file_content):
        """Determine coefficient version."""
        md5_hash = hashlib.md5(coeff_file_content)
        digest = md5_hash.hexdigest()
        version_dict = cls.version_hashs.get(
            digest,
            {"name": None, "status": None}
        )
        version = version_dict["name"]
        status = version_dict["status"]
        if version is None:
            warning = "Unknown calibration coefficients version!"
            warnings.warn(warning, RuntimeWarning)
            LOG.warning(warning)
        else:
            LOG.info('Identified calibration coefficients version "%s".',
                     version)
            if status != CoeffStatus.NOMINAL:
                warning = "Using {} calibration coefficients".format(status)
                warnings.warn(warning, RuntimeWarning)
                LOG.warning(warning)
        return version

    @staticmethod
    def date2float(date, decimals=5):
        """Convert date to year float.

        Argument
            date (datetime.datetime) - date
            decimals (int or None) - rounding precision if None, do not round (default=5)

        Return
            date_float (float) - date as year

        Note
            rounding to the 5th decimal reproduces the original float values from patmos-x

        Example:
            date2float('2000-07-02') == 2000.5
            because the 2nd of July was the middle day of the leap year 2000
        """
        year = date.year
        days_in_year = (dt.datetime(year+1, 1, 1) - dt.datetime(year, 1, 1)).days
        diff = date - dt.datetime(year, 1, 1)
        seconds = diff.total_seconds()
        date_float = date.year + seconds/(days_in_year*24*3600)
        if decimals is not None:
            date_float = round(date_float, decimals)
        return date_float


def calibrate_solar(counts, chan, year, jday, cal, corr=1):
    """Do the solar calibration and return scaled radiance.

    Arguments:
        counts (array) - raw counts for the given channels (options 1, 2[, 3A if available & active])
        chan (array) - pygac internal channel index array
        year (int) - year
        jday (int) - day of year
        cal (namedtuple) - spacecraft specific calibration coefficients, see Calibrator

    Optional:
        corr (float) - depricated - reflectance correction multiplier (default = 1)

    Returns:
        r_cal (array) - scaled radiance

    Note:
        This function and documentation follows the time-dependent solar calibration as described in:
        Heidinger, A.K., W.C. Straka III, C.C. Molling, J.T. Sullivan, and X. Wu, (2010).
        "Deriving an inter-sensor consistent calibration for the AVHRR solar reflectance data record.",
        International Journal of Remote Sensing, 31:6493 - 6517.
    """
    # Step 1. Obtain the calibration slope using equation (6) in Heidinger et al 2010
    # S(t) = S_0*(100 + S_1*t + S_2*t^2) / 100,
    # where t is the time since launch expressed as years, S(t) is the calibration slope
    # and S_0, S_1 and S_2 are the coefficients of the quadratic fit.
    # See also section "Fitting of Calibration Slope Equations" in PATMOS-x documentation
    # (CDRP-ATBD-0184 Rev. 2 03/29/2018 page 19)

    # Note that the this implementation does not take leap years into account!
    # Using datetime objects would include it automatically
    # sensing_date = dt.strptime('{0}.{1}'.format(year, jday), '%Y.%j')
    # delta = sensing_date - cal.date_of_launch
    # t = delta.total_seconds() / 31557600  # divided by seconds of a Julian year
    l_date = Calibrator.date2float(cal.date_of_launch)
    t = (year + jday / 365.0) - l_date

    # Note: splitting the calibration slope is needed to reproduce old results and may disappear in future,
    #       because actually there is only one set of slope parameters defined for single-gain counts
    #       as described in Heidinger et al. 2010. See Step 2 for more information.
    # Note that in case of a single-gain instrument, all gain_switch parameters are set to NaN.
    if np.isnan(cal.gain_switch).all():
        glow = ghigh = np.ones(3)
    else:
        glow = np.array([0.5, 0.5, 0.25])
        ghigh = np.array([1.5, 1.5, 1.75])
    # especially the rounding to three digits is crutial to exectly reproduce the original PATMOS-x values.
    al, bl, cl = np.round(glow*cal.s0, 3), cal.s1, cal.s2
    ah, bh, ch = np.round(ghigh*cal.s0, 3), cal.s1, cal.s2

    # apply slope equation for low and high gain coefficients
    stl = (al[chan] * (100.0 + bl[chan] * t + cl[chan] * t * t)) / 100.0
    sth = (ah[chan] * (100.0 + bh[chan] * t + ch[chan] * t * t)) / 100.0

    # Step 2. Calculate the scaled radiance using equation (1) in Heidinger et al 2010
    # R_cal = S*(C-D),
    # where R_cal is the value generated from the calibration and is referred to as a
    # scaled radiance, S is the calibration slope, C the measured count and D the dark count.
    # This equation is only valid for single-gain instruments. Starting with the AVHRR/3 series
    # (from NOAA-15 onwards), the channel-1, 2 and 3a require a dual-gain calibration.
    # The conversion for channel-1 and 2 is given in the appendix, equation (A1) and (A2).
    # In general, these equations can be written as
    # C(C_dg) = D + G_low*(C_dg-D),            if C_dg <= B_dg
    # C(C_dg) = C(B_dg) + G_high*(C_dg-B_dg),  otherwise
    # where C_dg is the measured dual-gain counts, B_dg is the dual-gain switch and G_low/high
    # are the gain factors for the low and high count region.
    # Quote from the book "Remote Sensing Time Series: Revealing Land Surface Dynamics":
    # > Another change in design of the AVHRR/3 instrument was the introduction of a dual-gain feature
    # > for the reflective channels 1, 2 and 3A. In order to improve the radiometric resolution of the
    # > instrument for low reflectance targets, the dynamic range of the instrument was divided equally
    # > in two ranges, i.e. nominally from 0 to 500 counts and from 500 to 1,000 counts. For channels 1
    # > and 2 half of the available Digital Number (DN) range is assigned to the low albedo range from
    # > 0 to 25% with the other half to the high albedo range from 26 to 100%. This allows for an increase
    # > in the radiometric resolution for dark targets. For channel 3A, the split between low and high
    # > albedo range is set at 12.5% albedo (Rao and Sullivan 2001).
    # The gain factors are given by the ratio of the fraction of albedo range to the fraction of count range
    # for the given count region. From the information given by the book quote, we get
    # G_low = 25% / 50% = 0.5 for channel-1 and 2
    # G_low = 12.5% / 50% = 0.25 for channel-3a
    # G_high = (100% - 25%) / (100% - 50%) = 1.5 for channel-1 and 2
    # G_high = (100% - 12.5%) / (100% - 50%) = 1.75 for channel-3a
    # Inserting the converted dual-gain counts into equation (1) yields the scaled radiance equation, which
    # is a continuous piecewise linear function of two line segments.
    #            R_cal
    #            ^           *         R_cal = S*G_low*(C_dg-D),                        if C_dg <= B_dg
    #            |          *          R_cal = S*G_low*(B_dg-D) + S*G_high*(C_dg-B_dg), otherwise
    #            |         *
    # R_cal(B_dg)|--------*
    #            |     *  |
    #          0 +--*-------------> C_dg
    #            *  D    B_dg
    # Note, that in the former implementation, there was a distinction beteen low and high gain slopes
    # given by S_low/high = S*G_low/high. which only affects S0 in equation (6) in Heidinger et al 2010.
    # Furthermore, the implementation allows for an additional correction factor corr which defaults to 1. (depricated)
    d = cal.dark_count[chan]
    b_dg = cal.gain_switch[chan]
    # Note that in case of a single-gain instrument, all gain_switch parameters are set to NaN.
    if not np.isnan(cal.gain_switch).all():
        r_cal = np.where(
            counts <= b_dg,
            (counts - d)*stl,
            (b_dg - d)*stl + (counts - b_dg)*sth
        )
    else:
        r_cal = stl*(counts - d)
    # apply depricated correction
    if corr != 1:
        warnings.warn(
            "Using the 'corr' argument is depricated in favor of making the units"
            " of the function result clear. Please make any unit conversion outside this function.",
            DeprecationWarning
        )
        r_cal *= corr

    # Mask negative scaled radiances
    r_cal[r_cal < 0] = np.nan

    return r_cal


def calibrate_thermal(counts, prt, ict, space, line_numbers, channel, cal):
    """Do the thermal calibration and return brightness temperatures (K).

    Arguments:
        counts (array) - counts for the given channel (options: 3B (if active), 4, 5)
        prt (array) - counts of the Platinum Resistance Thermometers (PRT)
        ict (array) - counts of the In-orbit Calibration Targets (ICT)
        space (array) - counts of cold space
        line_numbers (array) - line number index
        channel (array) - pygac internal channel index array
        cal (namedtuple) - spacecraft specific calibration coefficients, see Calibrator

    Note:
        This function and documentation follows steps 1 to 4 from the  KLM User's Guide
        (Robel, J. (2009). NOAA KLM user's guide with NOAA-N,-P supplement. NOAA KLM Users
        Guide - August 2014 Revision) section 7.1.2.4 "Steps to Calibrate the AVHRR Thermal Channels",
        and the smoothing approach by Trishchenko (2002).
        The correction method for the non-linear response of the Mercury-Cadmium-Telluride detectors used
        for channels 4 and 5 is based on Walton et al. (1998)
    """
    # Shift channel index by three to obtain thermal channels [3b, 4, 5].
    chan = channel - 3

    lines, columns = counts.shape[:2]

    # Step 1. The temperature of the internal blackbody target is measured by four platinum resistance
    # thermometers (PRT)s. In each scanline, data words 18, 19 and 20 in the HRPT minor frame format contain
    # three readings from one of the four PRTs. (See Section 4.1.3) A different PRT is sampled each scanline;
    # every fifth scanline all three PRT values are set equal to 0 to indicate that a set of four PRTs has
    # just been sampled. The count value CPRT of each PRT is converted to temperature TPRT by the formula
    # T_PRT = d0 + d1*C_PRT + d2*C_PRT^2 + d3*C_PRT^3 + d4*C_PRT^4    (7.1.2.4-1)
    # The coefficients d0, d1, d2, d3 and d4 vary slightly for each PRT. Values for the coefficients are
    # found in Appendix D, in Table D.1-8 for NOAA-15 (coefficients d3 and d4 are 0 for NOAA-15),
    # Table D.2-9 for NOAA-16, Table D.3-3 for NOAA-17 and Table D.4-3 for NOAA-18. To
    # calculate the internal blackbody temperature TBB, NESDIS uses the simple average
    # T_BB = (T_PRT1 + T_PRT2 + T_PRT3 + T_PRT4)/4    (7.1.2.4-2)

    # Find the corresponding PRT values for a given line number
    # Note that the prt values are the average value of the three readings from one of the four
    # PRTs. See reader.get_telemetry implementations.
    prt_threshold = 50  # empirically found and set by Abhay Devasthale

    for offset in range(5):
        # According to the KLM Guide the fill value between PRT measurments is 0, but we search
        # for the first measurement gap using the threshold, because the fill value is in practice
        # not always exactly 0.
        if np.median(prt[(line_numbers - line_numbers[0]) % 5 == offset]) < prt_threshold:
            break
    else:
        raise IndexError("No PRT 0-index found!")

    # get the PRT index, iprt equals to 0 corresponds to the measurement gaps
    iprt = (line_numbers - line_numbers[0] + 5 - offset) % 5

    # fill measured values below threshold by interpolation
    ifix = np.where(np.logical_and(iprt == 1, prt < prt_threshold))
    if len(ifix[0]):
        inofix = np.where(np.logical_and(iprt == 1, prt > prt_threshold))
        prt[ifix] = np.interp(ifix[0], inofix[0], prt[inofix])

    ifix = np.where(np.logical_and(iprt == 2, prt < prt_threshold))
    if len(ifix[0]):
        inofix = np.where(np.logical_and(iprt == 2, prt > prt_threshold))
        prt[ifix] = np.interp(ifix[0], inofix[0], prt[inofix])

    ifix = np.where(np.logical_and(iprt == 3, prt < prt_threshold))
    if len(ifix[0]):
        inofix = np.where(np.logical_and(iprt == 3, prt > prt_threshold))
        prt[ifix] = np.interp(ifix[0], inofix[0], prt[inofix])

    ifix = np.where(np.logical_and(iprt == 4, prt < prt_threshold))
    if len(ifix[0]):
        inofix = np.where(np.logical_and(iprt == 4, prt > prt_threshold))
        prt[ifix] = np.interp(ifix[0], inofix[0], prt[inofix])

    # calculate PRT temperature using equation (7.1.2.4-1) KLM Guide
    # Tprt = d0 + d1*Cprt + d2*Cprt^2 + d3*Cprt^3 + d4*Cprt^4
    # Note: First dimension of cal.d are the five coefficient indicees
    tprt = np.polynomial.polynomial.polyval(prt, cal.d[:, iprt], tensor=False)

    # Note: the KLM Guide proposes to calculate the mean temperature using equation (7.1.2.4-2).
    # PyGAC follows the smoothing approach by Trishchenko (2002), i.e.
    # filling the zeros that mark a complete set of thermometer measurements
    # by interpolation, and then using a weighting function (so far only equal
    # weighting) to convolve the temperatures to calculate a moving average of a given window size.
    # The same averaging technique is applied for ICTs and Space counts.
    zeros = iprt == 0
    nonzeros = np.logical_not(zeros)

    tprt[zeros] = np.interp((zeros).nonzero()[0],
                            (nonzeros).nonzero()[0],
                            tprt[nonzeros])

    # Thresholds to flag missing/wrong data for interpolation
    ict_threshold = 100
    space_threshold = 100
    if channel == 3:
        zeros = ict < ict_threshold
        nonzeros = np.logical_not(zeros)
        try:
            ict[zeros] = np.interp((zeros).nonzero()[0],
                                (nonzeros).nonzero()[0],
                                ict[nonzeros])
        except ValueError: # 3b has no valid data
            return counts
        zeros = space < space_threshold
        nonzeros = np.logical_not(zeros)

        space[zeros] = np.interp((zeros).nonzero()[0],
                                (nonzeros).nonzero()[0],
                                space[nonzeros])

    # convolving and smoothing PRT, ICT and SPACE values
    if lines > 51:
        wlength = 51  # empirically found and set by Abhay Devasthale
    else:
        wlength = 3

    weighting_function = np.ones(wlength, dtype=float) / wlength
    tprt_convolved = np.convolve(tprt, weighting_function, "same")
    ict_convolved = np.convolve(ict, weighting_function, "same")
    space_convolved = np.convolve(space, weighting_function, "same")

    # take care of the beginning and end
    tprt_convolved[0:(wlength - 1) // 2] = tprt_convolved[(wlength - 1) // 2]
    ict_convolved[0:(wlength - 1) // 2] = ict_convolved[(wlength - 1) // 2]
    space_convolved[0:(wlength - 1) // 2] = space_convolved[(wlength - 1) // 2]
    tprt_convolved[-(wlength - 1) // 2:] = tprt_convolved[-((wlength + 1) // 2)]
    ict_convolved[-(wlength - 1) // 2:] = ict_convolved[-((wlength + 1) // 2)]
    space_convolved[-(wlength - 1) // 2:] = space_convolved[-((wlength + 1) // 2)]

    new_tprt = np.transpose(np.tile(tprt_convolved, (columns, 1)))
    new_ict = np.transpose(np.tile(ict_convolved, (columns, 1)))
    new_space = np.transpose(np.tile(space_convolved, (columns, 1)))

    # Step 2. The radiance NBB sensed in each thermal AVHRR channel from the internal blackbody
    # at temperature TBB is the weighted mean of the Planck function over the spectral response of the
    # channel. [...]. Each thermal channel has one equation, which uses a centroid wavenumber labmda_c and an
    # "effective" blackbody temperature TBB*. The two steps are:
    # TsBB = A + B*TBB    (7.1.2.4-3)
    # NBB = c1*nu_e^3/(exp(c2*nu_e/TsBB) - 1)    (7.1.2.4-3)
    # where the constants of the Planck function are defined as c1 = 2*h*c^2, c2 = h*c/k_B.
    # constants
    c1 = 1.1910427e-5  # mW/m^2/sr/cm^{-4}
    c2 = 1.4387752  # cm K
    # coefficients
    A = cal.to_eff_blackbody_intercept[chan]
    B = cal.to_eff_blackbody_slope[chan]
    nu_c = cal.centroid_wavenumber[chan]
    nS = cal.space_radiance[chan]
    b = cal.b[chan, 0:3]  # the second index are the three polynomial coefficients
    # variables
    tBB = new_tprt
    cS = new_space
    cE = counts.astype(float)
    cBB = new_ict

    tsBB = A + B*tBB
    nBB_num = c1 * nu_c**3
    nBB = nBB_num / (np.exp((c2 * nu_c) / tsBB) - 1.0)

    # Step 3. Output from the two in-orbit calibration targets is used to compute a linear estimate of
    # the Earth scene radiance NE. Each scanline, the AVHRR views the internal blackbody target and
    # outputs 10 count values for each of the three thermal channel detectors; these are found in words
    # 23 to 52 in the HRPT data stream. When the AVHRR views cold space, 10 counts from each of
    # the five channel sensors are output and placed into words 52 to 102. (Table 4.1.3-1 describes
    # how these data are multiplexed.) Count values for each channel are averaged together to smooth
    # our random noise; often counts from five consecutive scanlines are averaged because it takes five
    # lines to obtain a set of all four PRT measurements. The average blackbody count CBB and the
    # average space count CS, together with blackbody radiance NBB and space radiance NS, explained
    # in the next paragraph, are used to compute the linear radiance estimate NLIN,
    # NLIN = NS + (NBB - NS)*(CS - CE)/(CS - CBB)    (7.1.2.4-5)
    # where CE is the AVHRR count output when it views one of the reference Earth targets.
    # While the detector in channel 3B has a linear response, the Mercury-Cadmium-Telluride detectors used
    # for channels 4 and 5 have a nonlinear response to incoming radiance. Pre-launch laboratory measurements show that:
    #     a. scene radiance is a slightly nonlinear (quadratic) function of AVHRR output count,
    #     b. the nonlinearity depends on the AVHRR operating temperature.
    # It is assumed that the nonlinear response will persist in orbit. For the NOAA KLM series of
    # satellites, NESDIS uses a radiance-based nonlinear correction method. In this method, the linear
    # radiance estimate is first computed using a non-zero radiance of space, the NS term in Equation
    # 7.1.2.4-5. Then, the linear radiance value is input into a quadratic equation to generate the
    # nonlinear radiance correction NCOR:
    # NCOR = b0 + b1*NLIN + b2*NLIN^2    (7.1.2.4-6)
    # Finally, the Earth scene radiance is obtained by adding NCOR to NLIN
    # NE = NLIN + NCOR

    # Note: For channel 3B, the non-linear correction coefficients are set to zero to use the same equation.
    Nlin = nS + (nBB - nS)*(cS - cE)/(cS - cBB)
    Ncor = np.polynomial.polynomial.polyval(Nlin, b[0:3], tensor=False)
    Ne = Nlin + Ncor

    # Step 4. Data users often convert the computed Earth scene radiance value NE into an equivalent
    # blackbody temperature TE. This temperature is defined by simple inverting the steps used to
    # calculate the radiance NE sensed by an AVHRR channel from an emitting blackbody at a
    # temperature TE. The two-step process is:
    # TsE = c2*nu_c / ln(1 + (c1*nu_c^3/NE))    (7.1.2.4-8)
    # TE = (TsE - A)/B    (7.1.2.4-9)
    tsE = c2*nu_c / np.log(1.0 + nBB_num / Ne)
    bt = (tsE - A) / B

    # Why do we do this on channel 3b?
    if chan == 0:
        bt = np.where((counts - new_space) >= 0, 0.0, bt)

    # Mask values outside valid range
    bt = np.where(np.logical_or(bt < 170.0, bt > 350.0), np.nan, bt)

    return bt



try:
    from pygac.version import version as __version__  # noqa
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "No module named pygac.version. This could mean "
        "you didn't install 'pygac' properly. Try reinstalling ('pip "
        "install').")

import logging

from pygac.configuration import get_config, read_config_file  # noqa
from pygac.runner import get_reader_class, process_file  # noqa

# add a NullHandler to prevent messages in sys.stderr if the using application does
# not use logging, but pygac makes logging calls of severity WARNING and greater.
# See https://docs.python.org/3/howto/logging.html (Configuring Logging for a Library)
logging.getLogger("pygac").addHandler(logging.NullHandler())
